from typing import Dict, List

from matplotlib import pyplot as plt
from .data_analysis import analyze_data
from pandas import DataFrame


# labels with units for the different properties
labels = {}
labels['temperature'] = 'Temperature (K)'
labels['potential_energy'] = 'Potential energy (eV)'
labels['Px'] = 'Pressure x (GPa)'
labels['Py'] = 'Pressure y (GPa)'
labels['Pz'] = 'Pressure z (GPa)'
labels['kx_tot'] = 'Kappa x (W/m K)'
labels['ky_tot'] = 'Kappa y (W/m K)'
labels['kz_tot'] = 'Kappa z (W/m K)'


def get_run_average(kappas: List[DataFrame],
                    nequil: int) -> dict:
    """Computes averages over several simulations and returns mean, error estimate
    and correlation length in the form of a dictionary.

    Parameters
    ----------
    kappas
        list of dataframes with data read from ``kappa.out`` files
    nequil
        number of data points to drop in the beginning to account for equilibration
    """
    df_avg = None
    for df in kappas.values():
        if df_avg is None:
            df_avg = df[df.index > nequil].copy()
        else:
            df_avg += df
    if len(kappas) == 0:
        return
    df_avg /= len(kappas)
    df_avg.dropna(inplace=True)

    tags = 'kx_tot ky_tot kz_tot'.split()
    data = {}
    for tag in tags:
        data[tag] = analyze_data(df_avg[tag])
    data = DataFrame.from_dict(data).T

    return data


def plot_thermos_split(thermos: Dict[str, DataFrame],
                       nsim_max: int = 10,
                       title: str = '') -> None:
    """Generates an overview figure of the thermodynamic data from a series of runs.

    Parameters
    ----------
    thermos
        dictionary with each entry containing a dataframe read from a ``thermo.out`` file
    nsim_max
        maximum number columns (=runs) to show
    title
        title string (optional)
    """
    tags = 'temperature potential_energy Px Py Pz'.split()
    nsim = min(nsim_max, len(thermos))
    fig, grid = plt.subplots(ncols=nsim, nrows=len(tags),
                             figsize=(12, 9), sharex='col', sharey='row')
    if len(title) > 0:
        fig.suptitle(title)
    for row, (tag, axes) in enumerate(zip(tags, grid)):
        axes[0].set_ylabel(f'{labels[tag]}')
        for ax, (run, df) in zip(axes, thermos.items()):
            ax.plot(df.index, df[tag])
            ax.set_xlabel('Index')
            if row == 0:
                ax.text(0.06, 0.92, run, transform=ax.transAxes)
    plt.tight_layout()
    fig.subplots_adjust(wspace=0, hspace=0)


def plot_kappas_split(kappas: Dict[str, DataFrame],
                      nsim_max: int = 10,
                      title: str = '') -> None:
    """Generates an overview figure of the thermal conductivity data from a series of runs.

    Parameters
    ----------
    kappas
        dictionary with each entry containing a dataframe read from a ``kappa.out``
        or ``hac.out`` file
    nsim_max
        maximum number columns (=runs) to show
    title
        title string (optional)
    """
    tags = 'kx_tot ky_tot kz_tot'.split()
    nsim = min(nsim_max, len(kappas))
    fig, grid = plt.subplots(ncols=nsim, nrows=len(tags),
                             figsize=(12, 6), sharex='col', sharey='row')
    if len(title) > 0:
        fig.suptitle(title)
    for row, (tag, axes) in enumerate(zip(tags, grid)):
        axes[0].set_ylabel(f'{labels[tag]}')
        for ax, (run, df) in zip(axes, kappas.items()):
            ax.plot(df.index, df[tag])
            ax.set_xlabel('Index')
            if row == 0:
                ax.text(0.06, 0.92, run, transform=ax.transAxes)
    plt.tight_layout()
    fig.subplots_adjust(wspace=0, hspace=0)


def plot_thermos_with_average(thermos: Dict[str, DataFrame],
                              title: str = '') -> None:
    """Generates an overview figure of the thermodynamic data from a series of runs.

    Parameters
    ----------
    kappas
        dictionary with each entry containing a dataframe read from a ``thermo.out`` file
    title
        title string (optional)
    """
    df_avg = None
    for df in thermos.values():
        if df_avg is None:
            df_avg = df.copy()
        else:
            df_avg += df
    if len(thermos) == 0:
        print('No data')
        return
    df_avg /= len(thermos)

    tags = 'temperature potential_energy Px Py Pz'.split()
    fig, axes = plt.subplots(ncols=len(tags), figsize=(12, 3), sharex=True)
    if len(title) > 0:
        fig.suptitle(title)
    for tag, ax in zip(tags, axes):
        ax.set_ylabel(f'{labels[tag]}')
        ax.set_xlabel('Index')
        for k, (run, df) in enumerate(thermos.items()):
            ax.plot(df.index, df[tag], color=f'{0.2+0.6*k/len(thermos)}')
        ax.plot(df_avg.index, df_avg[tag], color='red')
    plt.tight_layout()


def plot_kappas_with_average(kappas: Dict[str, DataFrame],
                             title: str = '') -> None:
    """Generates an overview figure of the thermal conductivity data from a series of runs.

    Parameters
    ----------
    kappas
        dictionary with each entry containing a dataframe read from a ``kappa.out``
        or ``hac.out`` file
    title
        title string (optional)
    """
    df_avg = None
    for df in kappas.values():
        if df_avg is None:
            df_avg = df.copy()
        else:
            df_avg += df
    if len(kappas) == 0:
        print('No data')
        return
    df_avg /= len(kappas)

    tags = 'kx_tot ky_tot kz_tot'.split()
    fig, axes = plt.subplots(ncols=len(tags), figsize=(9, 3), sharex=True)
    if len(title) > 0:
        fig.suptitle(title)
    for tag, ax in zip(tags, axes):
        ax.set_ylabel(f'{labels[tag]}')
        ax.set_xlabel('Index')
        for k, (run, df) in enumerate(kappas.items()):
            ax.plot(df.index, df[tag], color=f'{0.2+0.6*k/len(kappas)}')
        ax.plot(df_avg.index, df_avg[tag], color='red')
    plt.tight_layout()


def plot_kappas_distribution(kappas: Dict[str, DataFrame],
                             nequil: int = 0,
                             title: str = '') -> None:
    """Generates an overview figure of the thermal conductivity data from a series of runs.

    Parameters
    ----------
    kappas
        dictionary with each entry containing a dataframe read from a ``kappa.out`` file
    nequil
        number of steps to drop in the beginning of each trajectory to account for equilibration
    title
        title string (optional)
    """
    df_avg = None
    for df in kappas.values():
        if df_avg is None:
            df_avg = df[df.index > nequil].copy()
        else:
            df_avg += df
    if len(kappas) == 0:
        print('No data')
        return
    df_avg /= len(kappas)

    tags = 'kx_tot ky_tot kz_tot'.split()
    fig, axes = plt.subplots(ncols=len(tags), figsize=(9, 3), sharey=True)
    if len(title) > 0:
        fig.suptitle(title)
    for col, (tag, ax) in enumerate(zip(tags, axes)):
        if col == 0:
            ax.set_ylabel('Density')
        ax.set_xlabel(labels[tag])
        for k, (run, df) in enumerate(kappas.items()):
            df = df[df.index > nequil]
            try:
                ax.hist(df[tag], density=True, color=f'{0.2+0.6*k/len(kappas)}')
            except ValueError:
                print(f'Failed to generate histogram for {tag}')
        try:
            ax.hist(df_avg[tag], density=True, color='red')
        except ValueError:
            print('Failed to generate histogram for average')
    plt.tight_layout()
    fig.subplots_adjust(wspace=0, hspace=0)



# -*- coding: utf-8 -*-
"""
Main module of the calorine package.
"""

from .io import (read_kappa,
                 read_thermo,
                 read_settings,
                 read_xyz,
                 write_xyz)
from .nepy.cpunep_calculator import CPUNEP
from .gpunep_calculator import GPUNEP

__project__ = 'calorine'
__description__ = 'Some like it hot'
__copyright__ = '2022'
__license__ = 'Mozilla Public License 2.0 (MPL 2.0)'
__version__ = '1.1'
__maintainer__ = 'The calorine developers team'
__status__ = 'Stable'
__url__ = 'http://calorine.materialsmodeling.org/'

__all__ = ['CPUNEP',
           'GPUNEP',
           'read_thermo',
           'read_kappa',
           'read_settings',
           'read_xyz',
           'write_xyz']



from typing import Dict, List, Tuple

import numpy as np
from ase import Atoms
from ase.neighborlist import NeighborList
from pandas import DataFrame


def read_settings(filename: str) -> dict:
    """Returns the settings from a configuration file as a dictionary.

    Parameters
    ----------
    filename
        input file name
    """
    settings = {}
    if filename.endswith('run.in'):
        with open(filename) as f:
            for line in f.readlines():
                flds = line.split()
                if len(flds) == 0:
                    continue
                if flds[0].startswith('#'):
                    continue
                if flds[0] == 'compute_hnemd':
                    settings['dump_hnemd'] = int(flds[1])
                    settings['Fe_x'] = float(flds[2])
                    settings['Fe_y'] = float(flds[3])
                    settings['Fe_z'] = float(flds[4])
                elif flds[0] == 'compute_hac':
                    settings['sampling_interval'] = int(flds[1])
                    settings['correlation_steps'] = int(flds[2])
                    settings['output_interval'] = int(flds[3])
                elif flds[0] == 'ensemble':
                    settings['ensemble'] = flds[1]
                    if settings['ensemble'].startswith('nvt'):
                        settings['temperature'] = float(flds[2])
                        settings['temperature_final'] = float(flds[3])
                        settings['thermostat_coupling'] = float(flds[4])
                    elif settings['ensemble'].startswith('npt'):
                        settings['temperature'] = float(flds[2])
                        settings['temperature_final'] = float(flds[3])
                        settings['thermostat_coupling'] = float(flds[4])
                        settings['pressure_x'] = float(flds[5])
                        settings['pressure_y'] = float(flds[6])
                        settings['pressure_z'] = float(flds[7])
                        settings['pressure_coupling'] = float(flds[8])
                    elif settings['ensemble'].startswith('heat'):
                        settings['temperature'] = float(flds[2])
                        settings['thermostat_coupling'] = float(flds[3])
                        settings['delta_temperature'] = float(flds[4])
                        settings['label_source'] = flds[5]
                        settings['label_sink'] = flds[6]
                    else:
                        raise ValueError(f'Run.in contains invalid ensemble {settings["ensemble"]}.'
                                         f' Expected nvt, npt or heat.')
                elif flds[0] == 'time_step':
                    settings['time_step'] = float(flds[1])
                elif flds[0] == 'dump_thermo':
                    settings['dump_thermo'] = int(flds[1])
                elif flds[0] == 'neighbor':
                    settings['neighbor'] = int(flds[1])
                elif flds[0] == 'dump_position':
                    settings['dump_position'] = int(flds[1])
                elif flds[0] == 'velocity':
                    settings['velocity'] = float(flds[1])
                elif flds[0] == 'run':
                    settings['run'] = int(flds[1])
                elif flds[0] == 'potential':
                    settings[flds[0]] = ' '.join(flds[1:])
                    potential_filename = flds[1]
                    if 'fcp.txt' in potential_filename:
                        try:
                            with open(potential_filename, 'r') as f:
                                settings['fcp_order'] = int(f.readlines()[1].split()[0])
                        except IOError as e:
                            raise IOError(
                                f'Failed to read fcp potential in file {potential_filename}') from e
                else:
                    settings[flds[0]] = ' '.join(flds[1:])
    elif filename.endswith('nep.in'):
        with open(filename) as f:
            for line in f.readlines():
                flds = line.split()
                if len(flds) == 0:
                    continue
                if flds[0].startswith('#'):
                    continue
                settings[flds[0]] = ' '.join(flds[1:])
    else:
        raise ValueError(f'Unknown configuration file format for {filename}')
    return settings


def read_loss(filename: str) -> DataFrame:
    """Parses a file in ``loss.out`` format from GPUMD and returns the
    content as a data frame. More information concerning file format,
    content and units can be found at
    https://gpumd.zheyongfan.org/index.php/The_output_files_for_the_nep_executable#The_loss.out_file.

    Parameters
    ----------
    filename
        input file name

    """
    data = np.loadtxt(filename)
    if isinstance(data[0], np.float64):
        # If only a single row in loss.out, append a dimension
        data = data.reshape(1, -1)
    if len(data[0]) != 10:
        raise ValueError(f'Input file contains {len(data[0])} data columns.'
                         f' Expected 10 columns.')
    tags = 'total_loss L1 L2'
    tags += ' RMSE_E_train RMSE_F_train RMSE_V_train'
    tags += ' RMSE_E_test RMSE_F_test RMSE_V_test'
    generations = range(100, len(data) * 100 + 1, 100)
    df = DataFrame(data=data[:, 1:], columns=tags.split(), index=generations)
    return df


def read_nep(filename: str) -> Tuple[Dict[str, tuple], List[float]]:
    """Parses a file in ``nep.txt`` format from GPUMD and returns the
    content as a tuple of a dict and a list. The dict contains the
    information contained in the header, whereas the list contains the
    weights and rescaling factors.

    Parameters
    ----------
    filename
        input file name

    """
    header = []
    parameters = []
    nheader = 6
    with open(filename) as f:
        for k, line in enumerate(f.readlines()):
            flds = line.split()
            if len(flds) == 0:
                continue
            if k == 0 and 'zbl' in flds[0]:
                nheader += 1
            if k <= nheader:
                header.append(tuple(flds))
            elif len(flds) == 1:
                parameters.append(float(flds[0]))
            else:
                raise IOError(f'Failed to parse line {k} from {filename}')
    settings = {}
    for flds in header:
        if flds[0] in ['cutoff']:
            settings[flds[0]] = tuple(map(float, flds[1:]))
        elif flds[0] in ['zbl', 'n_max', 'l_max', 'ANN', 'basis_size']:
            settings[flds[0]] = tuple(map(int, flds[1:]))
        else:
            settings[flds[0]] = flds[1:]
    return settings, parameters


def read_kappa(filename: str) -> DataFrame:
    """Parses a file in ``kappa.out`` format from GPUMD and returns the
    content as a data frame. More information concerning file format,
    content and units can be found at
    https://gpumd.zheyongfan.org/index.php/The_kappa.out_output_file.

    Parameters
    ----------
    filename
        input file name

    """
    data = np.loadtxt(filename)
    if isinstance(data[0], np.float64):
        # If only a single row in kappa.out, append a dimension
        data = data.reshape(1, -1)
    tags = 'kx_in kx_out ky_in ky_out kz_tot'.split()
    if len(data[0]) != len(tags):
        raise ValueError(f'Input file contains {len(data[0])} data columns.'
                         f' Expected {len(tags)} columns.')
    df = DataFrame(data=data, columns=tags)
    df['kx_tot'] = df.kx_in + df.kx_out
    df['ky_tot'] = df.ky_in + df.ky_out
    return df


def read_hac(filename: str) -> DataFrame:
    """Parses a file in ``hac.out`` format from GPUMD and returns the
    content as a data frame. More information concerning file format,
    content and units can be found at
    https://gpumd.zheyongfan.org/index.php/The_hac.out_output_file.

    Parameters
    ----------
    filename
        input file name

    """
    data = np.loadtxt(filename)
    if isinstance(data[0], np.float64):
        # If only a single row in hac.out, append a dimension
        data = data.reshape(1, -1)
    tags = 'time'
    tags += ' jin_jtot_x jout_jtot_x jin_jtot_y jout_jtot_y jtot_jtot_z'
    tags += ' kx_in kx_out ky_in ky_out kz_tot'
    tags = tags.split()
    if len(data[0]) != len(tags):
        raise ValueError(f'Input file contains {len(data[0])} data columns.'
                         f' Expected {len(tags)} columns.')
    df = DataFrame(data=data, columns=tags)
    df['kx_tot'] = df.kx_in + df.kx_out
    df['ky_tot'] = df.ky_in + df.ky_out
    # remove columns with less relevant data to save space
    for col in df:
        if 'jtot' in col or '_in' in col:
            del df[col]
    return df


def read_thermo(filename: str,
                natoms: int = 1) -> DataFrame:
    """Parses a file in ``thermo.out`` format from GPUMD and returns the
    content as a data frame. More information concerning file format,
    content and units can be found at
    https://gpumd.zheyongfan.org/index.php/The_thermo.out_output_file.

    Parameters
    ----------
    filename
        input file name
    natoms
        number of atoms; used to normalize energies

    """
    data = np.loadtxt(filename)
    if isinstance(data[0], np.float64):
        # If only a single row in loss.out, append a dimension
        data = data.reshape(1, -1)
    if len(data[0]) == 9:
        # orthorhombic box
        tags = 'temperature kinetic_energy potential_energy'
        tags += ' stress_xx stress_yy stress_zz'
        tags += ' cell_xx cell_yy cell_zz'
    elif len(data[0]) == 12:
        # orthorhombic box with stresses in Voigt notation (v3.3.1+)
        tags = 'temperature kinetic_energy potential_energy'
        tags += ' stress_xx stress_yy stress_zz stress_yz stress_xz stress_xy'
        tags += ' cell_xx cell_yy cell_zz'
    elif len(data[0]) == 15:
        # triclinic box
        tags = 'temperature kinetic_energy potential_energy'
        tags += ' stress_xx stress_yy stress_zz'
        tags += ' cell_xx cell_xy cell_xz cell_yx cell_yy cell_yz cell_zx cell_zy cell_zz'
    elif len(data[0]) == 18:
        # triclinic box with stresses in Voigt notation (v3.3.1+)
        tags = 'temperature kinetic_energy potential_energy'
        tags += ' stress_xx stress_yy stress_zz stress_yz stress_xz stress_xy'
        tags += ' cell_xx cell_xy cell_xz cell_yx cell_yy cell_yz cell_zx cell_zy cell_zz'
    else:
        raise ValueError(f'Input file contains {len(data[0])} data columns.'
                         ' Expected 9, 12, 15 or 18 columns.')
    df = DataFrame(data=data, columns=tags.split())
    assert natoms > 0, 'natoms must be positive'
    df.kinetic_energy /= natoms
    df.potential_energy /= natoms
    return df


def read_xyz(filename: str) -> Tuple[Atoms, Dict[str, int]]:
    """
    Read the structure input file (`xyz.in`) for GPUMD and return the
    structure along with run input parameters from the file.

    Parameters
    ----------
    filename
        Name of file from which to read the structure

    Returns
    -------
    tuple comprising the structure and the parameters from the first row of the file
    """

    with open(filename, 'r') as fd:

        # Parse first line
        first_line = next(fd)
        input_parameters = {}
        keys = ['number_of_atoms', 'maximum_neighbors', 'cutoff',
                'use_triclinic', 'has_velocity', 'number_of_groups']
        types = [float if key == 'cutoff' else int for key in keys]
        for k, (key, typ) in enumerate(zip(keys, types)):
            input_parameters[key] = typ(first_line.split()[k])

        # Parse second line
        second_line = next(fd)
        second_arr = np.array(second_line.split())
        pbc = second_arr[:3].astype(bool)
        if input_parameters['use_triclinic']:
            cell = second_arr[3:].astype(float).reshape((3, 3))
        else:
            cell = np.diag(second_arr[3:].astype(float))

        # Parse all remaining rows
        n_rows = input_parameters['number_of_atoms']
        n_columns = 5 + input_parameters['has_velocity'] * 3
        n_columns += input_parameters['number_of_groups']
        rest_lines = [next(fd) for _ in range(n_rows)]
        rest_arr = np.array([line.split() for line in rest_lines])
        assert rest_arr.shape == (n_rows, n_columns)

    # Extract atom types, positions and masses
    symbols = rest_arr[:, 0].astype('str')
    positions = rest_arr[:, 1:4].astype(float)

    # Create the Atoms object
    structure = Atoms(symbols=symbols, positions=positions, pbc=pbc, cell=cell)
    if input_parameters['has_velocity']:
        velocities = rest_arr[:, 5:8].astype(float)
        structure.set_velocities(velocities)
    if input_parameters['number_of_groups']:
        start_col = 5 + 3 * input_parameters['has_velocity']
        groups = rest_arr[:, start_col:].astype(int)
        structure.info = {i: {'groups': groups[i, :]} for i in range(n_rows)}

    return structure, input_parameters


def write_xyz(filename: str,
              structure: Atoms,
              maximum_neighbors: int = None,
              cutoff: float = None,
              groupings: List[List[List[int]]] = None,
              use_triclinic: bool = False):
    """
    Writes a structure into GPUMD input format (`xyz.in`).

    Parameters
    ----------
    filename
        Name of file to which the structure should be written
    structure
        Input structure
    maximum_neighbors
        Maximum number of neighbors any atom can ever have (not relevant when
        using force constant potentials)
    cutoff
        Initial cutoff distance used for building the neighbor list (not
        relevant when using force constant potentials)
    groupings
        Groups into which the individual atoms should be divided in the form of
        a list of list of lists. Specifically, the outer list corresponds to
        the grouping methods, of which there can be three at the most, which
        contains a list of groups in the form of lists of site indices. The
        sum of the lengths of the latter must be the same as the total number
        of atoms.
    use_triclinic
        Use format for triclinic cells

    Raises
    ------
    ValueError
        Raised if parameters are incompatible
    """

    # Check velocties parameter
    velocities = structure.get_velocities()
    if velocities is None or np.max(np.abs(velocities)) < 1e-6:
        has_velocity = 0
    else:
        has_velocity = 1

    # Check groupings parameter
    if groupings is None:
        number_of_grouping_methods = 0
    else:
        number_of_grouping_methods = len(groupings)
        if number_of_grouping_methods > 3:
            raise ValueError('There can be no more than 3 grouping methods!')
        for g, grouping in enumerate(groupings):
            all_indices = [i for group in grouping for i in group]
            if len(all_indices) != len(structure) or \
                    set(all_indices) != set(range(len(structure))):
                raise ValueError(f'The indices listed in grouping method {g} are'
                                 ' not compatible with the input'
                                 ' structure!')

    # If not specified, estimate the maximum_neighbors
    if maximum_neighbors is None:
        if cutoff is None:
            cutoff = 0.1
            maximum_neighbors = 1
        else:
            nl = NeighborList([cutoff / 2] * len(structure), skin=2, bothways=True)
            nl.update(structure)
            maximum_neighbors = 0
            for atom in structure:
                maximum_neighbors = max(maximum_neighbors,
                                        len(nl.get_neighbors(atom.index)[0]))
            maximum_neighbors *= 2
            maximum_neighbors = min(maximum_neighbors, 1024)

    # Add header and cell parameters
    lines = []
    if structure.cell.orthorhombic and not use_triclinic:
        triclinic = 0
    else:
        triclinic = 1
    lines.append('{} {} {} {} {} {}'.format(len(structure), maximum_neighbors,
                                            cutoff, triclinic, has_velocity,
                                            number_of_grouping_methods))
    if triclinic:
        lines.append((' {}' * 12)[1:].format(*structure.pbc.astype(int),
                                             *structure.cell[:].flatten()))
    else:
        lines.append((' {}' * 6)[1:].format(*structure.pbc.astype(int),
                                            *structure.cell.lengths()))

    # Add lines for all atoms
    for a, atm in enumerate(structure):
        line = (' {}' * 5)[1:].format(atm.symbol, *atm.position, atm.mass)
        if has_velocity:
            line += (' {}' * 3).format(*velocities[a])
        if groupings is not None:
            for grouping in groupings:
                for i, group in enumerate(grouping):
                    if a in group:
                        line += ' {}'.format(i)
        lines.append(line)

    # Write file
    with open(filename, 'w') as fd:
        fd.write('\n'.join(lines))


# Deprecated! With the `dump_exyz` command one gets `dump.xyz` files
# that contain positions, velocities and forces
# def _generator_vf(fname: str,
#                   n_atoms: int):
#     """
#     Generator for velocity and force data from GPUMD
#     """
#     with open(fname) as f:
#         while True:
#             data = np.loadtxt(f, max_rows=n_atoms, dtype=float)
#             if len(data) == 0:
#                 break
#             yield data


# def convert_gpumd_run_to_traj(dirname: str,
#                               traj_fname: str,
#                               include_velocities: bool = False,
#                               include_forces: bool = False):
#     """Convert a gpumd trajectory (movie.xyz, velocity.out, force.out) to
#     an ASE trajectory file.

#     Parameters
#     ----------
#     dirname
#         gpumd run directory
#     traj_fname
#         output trajectory filename

#     Example
#     -------
#     >>> dirname = 'gpumd-run'
#     >>> traj_fname = 'traj_T100.traj'
#     >>> convert_gpumd_run_to_traj(dirname, traj_fname, include_velocities=True)
#     """

#     # picoseconds to (amu / eV)**0.5 * Ang
#     ps_to_ase = 1e-12 * units.second

#     # setup trajectory file
#     traj = Trajectory(traj_fname, mode='w')

#     # filenames
#     fname_xyz = os.path.join(dirname, 'movie.xyz')
#     fname_v = os.path.join(dirname, 'velocity.out')
#     fname_f = os.path.join(dirname, 'force.out')
#     if not os.path.isfile(fname_xyz):
#         raise ValueError(f'{fname_xyz} does not exist')

#     # read first line to get number of atoms
#     fobj = open(fname_xyz, 'r')
#     n_atoms = int(fobj.readline())
#     print('natoms', n_atoms)
#     fobj.seek(0)

#     # setup generators
#     indices = slice(None, None, 1)
#     gen_xyz = iread_xyz(fobj, indices=indices)
#     gen_v = _generator_vf(fname_v, n_atoms)
#     gen_f = _generator_vf(fname_f, n_atoms)

#     # loop over all snapshots
#     snapshot_index = 0
#     while True:
#         if snapshot_index % 100 == 0:
#             print('  snapshot', snapshot_index)

#         # get xyz
#         try:
#             atoms = next(gen_xyz)
#         except Exception:
#             break
#         assert len(atoms) == n_atoms

#         # get velocities
#         if include_velocities:
#             v = next(gen_v)
#             assert v.shape == (n_atoms, 3), 'Velocities have wrong shape!'
#             atoms.set_velocities(v / ps_to_ase)

#         # get forces
#         if include_forces:
#             f = next(gen_f)
#             assert f.shape == (n_atoms, 3), 'Forces have wrong shape!'
#             calc = SinglePointCalculator(atoms, forces=f)
#             atoms.calc = calc

#         # finalize and write
#         snapshot_index += 1
#         traj.write(atoms)

#     fobj.close()
#     print(f'Converted trajectory of length {snapshot_index}')


# Deprecated! Demonstrate in examples
# def update_database(db_file: str,
#                     runpath: str,
#                     overwrite: bool = False,
#                     ndirs: int = 1,
#                     config_filename: str = 'run.in') -> int:
#     """Parses the runs in the given run directory and compiles the
#     information into an entry of the database. The index of the entry
#     is returned. It is assumed that there is a series of
#     subdirectories in ``runpath``, which each contain a ``kappa.out``
#     or ``hac.out`` and a ``thermo.out`` file. The ``runpath`` should
#     contain a configuration file (``run.in`` or ``config``) and a
#     structure file (``xyz.in``).

#     The last part of the run directory name is used as a unique
#     identifier. If an entry with the name already exists it will be
#     updated. If not a new row is added.

#     Parameters
#     ----------
#     db_file
#         path to database file
#     runpath
#         path to main run directory
#     overwrite
#         if True overwrite ancentry with the same directory name;
#         if False skip the entry
#     ndirs
#         number of directory levels that are used to generate the dirname field
#     config_filename
#         name of configuration; possible values: `run.in`, `config`

#     """
#     dirname = '/'.join(os.path.normpath(runpath).split('/')[-ndirs:])

#     with connect(db_file) as db:
#         try:
#             row = db.get(dirname=dirname)
#             index = row.id
#         except KeyError:
#             index = None
#         if index is not None and not overwrite:
#             print(f'Skipping row {index}; dirname: {dirname}')
#             return

#     # compile information
#     settings = read_settings(os.path.join(runpath, config_filename))
#     structure = read_xyz(os.path.join(runpath, 'xyz.in'))
#     structure.calc = SinglePointCalculator(structure)
#     structure.calc.name = 'gpumd'
#     structure.calc.parameters = settings

#     data = {}

#     kappas = {}
#     for fname in sorted(glob(os.path.join(runpath, '*/kappa.out'))):
#         tag = os.path.dirname(fname).split('/')[-1]
#         kappas[tag] = read_thermo(fname)
#     if len(kappas):
#         data['kappas'] = {key: {col: series.to_numpy() for col, series in df.iteritems()}
#                           for key, df in kappas.items()}

#     hacs = {}
#     for fname in sorted(glob(os.path.join(runpath, '*/hac.out'))):
#         tag = os.path.dirname(fname).split('/')[-1]
#         hacs[tag] = read_thermo(fname)
#     if len(hacs):
#         data['hacs'] = {key: {col: series.to_numpy() for col, series in df.iteritems()}
#                         for key, df in hacs.items()}

#     thermos = {}
#     for fname in sorted(glob(os.path.join(runpath, '*/thermo.out'))):
#         tag = os.path.dirname(fname).split('/')[-1]
#         thermos[tag] = read_thermo(fname)
#     if len(thermos):
#         data['thermos'] = {key: {col: series.to_numpy() for col, series in df.iteritems()}
#                            for key, df in thermos.items()}

#     # compile data that will be stored in fields
#     hostname = socket.gethostname()
#     username = getpass.getuser()
#     information = {'dirname': dirname,
#                    'hostname': hostname,
#                    'username': username,
#                    'data': data,
#                    'n_runs': len(data['thermos'])}
#     for tag in ['temperature', 'Fe_x', 'Fe_y', 'Fe_z', 'fcp_order']:
#         if tag in settings:
#             information[tag] = settings[tag]

#     # add to or update database
#     with connect(db_file) as db:
#         try:
#             row = db.get(dirname=dirname)
#             index = row.id
#         except KeyError:
#             index = None

#         if index is None:
#             print(f'Adding row; dirname: {dirname}')
#             index = db.write(structure, **information)
#         else:
#             print(f'Updating row {index}; dirname: {dirname}')
#             db.update(index, structure, **information)

#     return index


# def read_from_database(row: AtomsRow) -> Tuple[dict, Dict[str, DataFrame], Dict[str, DataFrame]]:
#     """Parses a database entry and returns a tuple comprising the run
#     settings, the thermal conductivity data (from ``kappa.out`` or
#     ``hac.out`` files) and the thermodynamic data (from ``thermo.out`` files).

#     Parameters
#     ----------
#     row
#         row from ASE database

#     """
#     thermos = {}
#     for key, data in row.data['thermos'].items():
#         thermos[key] = DataFrame.from_dict(data)
#         kp_data = {}
#     if 'kappas' in row.data:
#         for key, data in row.data['kappas'].items():
#             kp_data[key] = DataFrame.from_dict(data)
#     elif 'hacs' in row.data:
#         for key, data in row.data['hacs'].items():
#             kp_data[key] = DataFrame.from_dict(data)
#     return row.calculator_parameters, kp_data, thermos



import pytest
import numpy as np
from ase import Atoms
from calorine.nepy.nep import get_descriptors, get_potential_forces_and_virials, _clean_tmp_dir

PbTe = Atoms('TePb', positions=[(0, 0, 0), (0, 0.0, 1.1)],
             cell=([100, 0, 0], [0, 100, 0], [0, 0, 100]))
C = Atoms('C', positions=[(0, 0, 0)])
CO = Atoms('CO', positions=[(0, 0, 0), (0, 0.0, 1.1)])
CC = Atoms('CC', positions=[(0, 0, 0), (0, 0.0, 1.1)])
CON = Atoms('CON', positions=[(0, 0, 0), (0, 0.0, 1.1), (0, 0.0, 2.2)])


def _load_nep_from_file(file: str) -> str:
    """Load a NEP model from file into a stringified version

    Parameters
    ----------
    file
        Path to nep.txt.

    Returns
    -------
    str
        Stringified NEP model.
    """
    with open(file, 'r') as f:
        model = f.read()
    return model


# --- get_descriptors ---
def test_get_descriptors_setup_dummy_NEP2_model():
    """Verifies the dummy NEP model is properly formatted."""
    get_descriptors(PbTe, debug=True)
    tmp_dir = './tmp_nepy/'
    PbTe_dummy = _load_nep_from_file(f'{tmp_dir}/nep.txt')
    _clean_tmp_dir(tmp_dir)
    expected_PbTe_dummy = _load_nep_from_file(
        'tests/nep_models/PbTe_NEP2_dummy.txt')
    assert PbTe_dummy == expected_PbTe_dummy


def test_get_descriptors_no_cell():
    """Should get descriptors for atoms without a specified cell, and raise a warning."""
    with pytest.warns(UserWarning) as record:
        descriptors = get_descriptors(CO)
    assert len(record) == 1
    assert record[0].message.args[0] == 'Using default unit cell (cubic with side 100 Å).'
    assert descriptors.shape == (2, 52)


def test_get_descriptors_NEP2_independent_of_species():
    """NEP2 should give the same descriptors regardless of atom species."""
    descriptors_PbTe = get_descriptors(PbTe)
    descriptors_CO = get_descriptors(CO)
    assert np.allclose(descriptors_CO, descriptors_PbTe, atol=1e-12, rtol=0)


def test_get_descriptors_NEP2_several_atoms_same_species():
    """NEP2 should get the descritors for a single component system"""
    descriptors_CC = get_descriptors(CC)
    assert descriptors_CC.shape == (2, 52)
    assert descriptors_CC.dtype == np.float64
    assert not np.all(np.isclose(descriptors_CC, 0))


def test_get_descriptors_dummy_NEP2_several_atom_species():
    """Verifies the dummy NEP model has the correct number of parameters."""
    # C
    get_descriptors(C, debug=True)
    tmp_dir = './tmp_nepy/'
    C_dummy_parameters = _load_nep_from_file(f'{tmp_dir}/nep.txt').split('\n')[6:]
    _clean_tmp_dir(tmp_dir)
    assert len(C_dummy_parameters) == 1698
    # CO
    get_descriptors(CO, debug=True)
    tmp_dir = './tmp_nepy/'
    CO_dummy_parameters = _load_nep_from_file(f'{tmp_dir}/nep.txt').split('\n')[6:]
    _clean_tmp_dir(tmp_dir)
    assert len(CO_dummy_parameters) == 1773
    # CON
    get_descriptors(CON, debug=True)
    tmp_dir = './tmp_nepy/'
    CON_dummy_parameters = _load_nep_from_file(f'{tmp_dir}/nep.txt').split('\n')[6:]
    _clean_tmp_dir(tmp_dir)
    assert len(CON_dummy_parameters) == 1898


def test_get_descriptors_NEP2_dummy():
    """Case: No NEP model supplied; using dummy NEP2 model.
       Compares results to output from `nep_cpu`
    """
    descriptors_PbTe = get_descriptors(PbTe)
    expected_PbTe = np.loadtxt(
        'tests/expected_descriptors/PbTe_2atom_NEP2_dummy.out')
    assert np.allclose(descriptors_PbTe, expected_PbTe, atol=1e-12, rtol=0)


def test_get_descriptors_NEP3():
    """Case: NEP3 model supplied. Compares results to output from `nep_cpu`"""
    nep3 = 'tests/nep_models/PbTe_NEP3.txt'
    descriptors_PbTe = get_descriptors(PbTe, potential_filename=nep3)
    expected_PbTe = np.loadtxt(
        'tests/expected_descriptors/PbTe_2atom_NEP3.out')
    assert np.allclose(descriptors_PbTe, expected_PbTe, atol=1e-12, rtol=0)


def test_get_descriptors_debug():
    """Check that the generated files are accessible in the debug directory"""
    get_descriptors(PbTe, debug=True)
    tmp_dir = './tmp_nepy/'
    PbTe_dummy = _load_nep_from_file(f'{tmp_dir}/nep.txt')
    _clean_tmp_dir(tmp_dir)
    assert 'nep 2 Te Pb' in PbTe_dummy


def test_get_descriptors_debug_directory_exists():
    """Should fail if debug directory already exists from an earlier calculation"""
    get_descriptors(PbTe, debug=True)
    tmp_dir = './tmp_nepy/'
    with pytest.raises(FileExistsError)as e:
        get_descriptors(PbTe, debug=True)
    assert 'Please delete or move the conflicting directory' in str(e)
    _clean_tmp_dir(tmp_dir)


# --- get_potential_forces_and_virials ---
def test_get_potential_forces_and_virials_NEP3():
    """Case: NEP3 model supplied. Compares results to output from `nep_cpu`"""
    nep3 = 'tests/nep_models/PbTe_NEP3.txt'
    energies, forces, virials = get_potential_forces_and_virials(
        PbTe, potential_filename=nep3)
    assert energies.shape == (2,)
    assert forces.shape == (2, 3)
    assert virials.shape == (2, 9)
    expected_forces = np.loadtxt(
        'tests/expected_descriptors/PbTe_2atom_NEP3_force.out')
    assert np.allclose(forces, expected_forces, atol=1e-12, rtol=0)


def test_get_potential_forces_and_virials_NEP3_debug():
    """Compares result with debug flag enabled."""
    nep3 = 'tests/nep_models/PbTe_NEP3.txt'
    energies, forces, virials = get_potential_forces_and_virials(
        PbTe, potential_filename=nep3, debug=True)
    assert energies.shape == (2,)
    assert forces.shape == (2, 3)
    assert virials.shape == (2, 9)
    expected_forces = np.loadtxt(
        'tests/expected_descriptors/PbTe_2atom_NEP3_force.out')
    assert np.allclose(forces, expected_forces, atol=1e-12, rtol=0)


def test_get_potential_forces_and_virials_dummy_NEP2():
    """Dummy NEP2 model supplied. Compares results to output from `nep_cpu` for another system"""
    nep3 = 'tests/nep_models/CO_NEP2_dummy.txt'
    energies, forces, virials = get_potential_forces_and_virials(
        CO, potential_filename=nep3)
    assert energies.shape == (2,)
    assert forces.shape == (2, 3)
    assert virials.shape == (2, 9)
    expected_forces = np.loadtxt(
        'tests/expected_descriptors/CO_2atom_NEP2_dummy_force.out')
    assert np.allclose(forces, expected_forces, atol=1e-12, rtol=0)


def test_get_potential_forces_and_virials_no_cell():
    """Should work with default cell if no cell is supplied"""
    nep2_dummy = 'tests/nep_models/CO_NEP2_dummy.txt'
    with pytest.warns(UserWarning) as record:
        energies, forces, virials = get_potential_forces_and_virials(
            CO, potential_filename=nep2_dummy)
    assert len(record) == 1
    assert record[0].message.args[0] == 'Using default unit cell (cubic with side 100 Å).'
    assert energies.shape == (2,)
    assert forces.shape == (2, 3)
    assert virials.shape == (2, 9)
    expected_forces = np.loadtxt(
        'tests/expected_descriptors/CO_2atom_NEP2_dummy_force.out')
    assert np.allclose(forces, expected_forces, atol=1e-12, rtol=0)


def test_get_potential_forces_and_virials_several_of_same_species():
    """Check that forces are correct for a CC system"""
    nep = 'tests/nep_models/C_NEP2_dummy.txt'
    energies, forces, virials = get_potential_forces_and_virials(
        CC, potential_filename=nep)
    assert energies.shape == (2,)
    assert forces.shape == (2, 3)
    assert virials.shape == (2, 9)
    expected_forces_PbTe = np.loadtxt(
        'tests/expected_descriptors/C_2atom_NEP2_dummy_force.out')
    assert np.allclose(forces, expected_forces_PbTe, atol=1e-12, rtol=0)


def test_get_potential_forces_and_virials_several_different_species():
    """Check that forces are correct for a CON system.
    Note that these forces should be exactly zero for this system
    since the NEP2 dummy potential treats all atom species as identical atm.
    """
    nep = 'tests/nep_models/CON_NEP2_dummy.txt'
    energies, forces, virials = get_potential_forces_and_virials(
        CON, potential_filename=nep)
    assert energies.shape == (3,)
    assert forces.shape == (3, 3)
    assert virials.shape == (3, 9)
    expected_forces_PbTe = np.loadtxt(
        'tests/expected_descriptors/CON_3atom_NEP2_dummy_force.out')
    assert np.allclose(forces, expected_forces_PbTe, atol=1e-12, rtol=0)
    assert np.allclose(forces, np.zeros((3, 3)), atol=1e-12, rtol=0)


def test_get_potential_forces_and_virials_no_potential():
    """Tries to get potentials, forces and virials without specifying potential"""
    with pytest.raises(ValueError)as e:
        get_potential_forces_and_virials(PbTe)
    assert 'Potential must be defined!' in str(e)



import tempfile
import pytest

import numpy as np
from ase import Atoms
from ase.io import read, write
from ase.build import bulk
from calorine.io import read_settings, read_loss, read_nep, \
    read_thermo, read_kappa, read_hac, read_xyz, write_xyz


# --- read_settings ---
def test_read_settings_run_nvt():
    """Reads a run.in file with an NVT ensemble"""
    settings = read_settings('tests/example_files/run.in')
    assert settings['dump_hnemd'] == 1
    assert settings['Fe_x'] == 2
    assert settings['Fe_y'] == 3
    assert settings['Fe_z'] == 4

    assert settings['sampling_interval'] == 100
    assert settings['correlation_steps'] == 666
    assert settings['output_interval'] == 200

    assert settings['ensemble'] == 'nvt'
    assert settings['temperature'] == 300
    assert settings['temperature_final'] == 400
    assert settings['thermostat_coupling'] == 1

    assert settings['time_step'] == 1
    assert settings['dump_thermo'] == 2
    assert settings['neighbor'] == 66
    assert settings['dump_position'] == 3
    assert settings['velocity'] == 123
    assert settings['run'] == 1337


def test_read_settings_run_npt(tmpdir):
    """Reads a run.in file with an NPT ensemble"""
    p = tmpdir.join('run.in')
    p.write('ensemble npt 300 400 1 100 200 300 4\n')
    settings = read_settings(f'{p}')
    assert settings['ensemble'] == 'npt'
    assert settings['temperature'] == 300
    assert settings['temperature_final'] == 400
    assert settings['thermostat_coupling'] == 1
    assert settings['pressure_x'] == 100
    assert settings['pressure_y'] == 200
    assert settings['pressure_z'] == 300
    assert settings['pressure_coupling'] == 4


def test_read_settings_run_heat(tmpdir):
    """Reads a run.in file for heat transport"""
    p = tmpdir.join('run.in')
    p.write('ensemble heat 300 1 20 A B\ntime_step 0.1\n')
    settings = read_settings(f'{p}')
    assert settings['ensemble'] == 'heat'
    assert settings['temperature'] == 300
    assert settings['thermostat_coupling'] == 1
    assert settings['delta_temperature'] == 20
    assert settings['label_source'] == 'A'
    assert settings['label_sink'] == 'B'
    assert settings['time_step'] == 0.1


def test_read_settings_fcp_potential(tmpdir):
    """Reads a run.in file with a fcp potential"""
    p = tmpdir.join('run.in')
    p.write('potential tests/example_files/fcp/fcp.txt\n')
    settings = read_settings(f'{p}')
    assert settings['fcp_order'] == 2


def test_read_settings_run_nep_potential(tmpdir):
    """Reads a run.in file with a nep potential. At the moment only saves the filename."""
    p = tmpdir.join('run.in')
    p.write('potential tests/example_files/nep.txt\n')
    settings = read_settings(f'{p}')
    assert settings['potential'] == 'tests/example_files/nep.txt'


def test_read_settings_run_invalid_potential(tmpdir):
    """Tries to read a run.in file with an invalid potential"""
    p = tmpdir.join('run.in')
    run_invalid = tmpdir.join('fcp.txt')
    p.write(f'potential {run_invalid}\n')
    with pytest.raises(IOError) as e:
        read_settings(f'{p}')
    assert f'Failed to read fcp potential in file {run_invalid}' in str(e)


def test_read_settings_unknown_setting(tmpdir):
    """Reads a run.in file with a comment"""
    p = tmpdir.join('run.in')
    p.write('test 1\n')
    settings = read_settings(f'{p}')
    assert settings['test'] == '1'


def test_read_settings_run_comment(tmpdir):
    """Reads a run.in file with a comment"""
    p = tmpdir.join('run.in')
    p.write('#ensemble heat 300 1 20 A B\n')
    settings = read_settings(f'{p}')
    assert settings == {}


def test_read_settings_run_blank_line(tmpdir):
    """Reads a run.in file with a blank line"""
    p = tmpdir.join('run.in')
    p.write('\n')
    settings = read_settings(f'{p}')
    assert settings == {}


def test_read_settings_invalid_ensemble(tmpdir):
    """Tries to read a run.in with invalid ensemble"""
    p = tmpdir.join('run.in')
    p.write('ensemble hmm\n')
    with pytest.raises(ValueError) as e:
        read_settings(f'{p}')
    error_string = 'Run.in contains invalid ensemble hmm. '\
        'Expected nvt, npt or heat.'
    assert error_string in str(e)


def test_read_settings_nep():
    """Reads a nep.in file"""
    settings = read_settings('tests/example_files/nep.in')
    assert settings['version'] == '4'
    assert settings['type'] == '2 C H'
    assert settings['cutoff'] == '8 4'
    assert settings['n_max'] == '8 6'
    assert settings['l_max'] == '4'
    assert settings['neuron'] == '50'
    assert settings['lambda_1'] == '0.01'
    assert settings['lambda_2'] == '0.01'
    assert settings['batch'] == '100'
    assert settings['population'] == '50'
    assert settings['generation'] == '50000'
    assert settings['lambda_e'] == '1'
    assert 'lambda_f' not in settings.keys()


def test_read_settings_nep_blank_line(tmpdir):
    """Reads a nep.in file with a blank line"""
    p = tmpdir.join('nep.in')
    p.write('\n')
    settings = read_settings(f'{p}')
    assert settings == {}


def test_read_settings_unknown_format(tmpdir):
    """Tries to read an unknown file format"""
    p = tmpdir.join('run_invalid.in')
    with pytest.raises(ValueError) as e:
        read_settings(f'{p}')
    assert f'Unknown configuration file format for {p}' in str(e)


# --- read_loss ---
def test_read_loss():
    """Reads a loss.out file"""
    loss = read_loss('tests/example_files/loss.out')
    columns_check = loss.columns == ['total_loss', 'L1', 'L2',
                                     'RMSE_E_train', 'RMSE_F_train', 'RMSE_V_train',
                                     'RMSE_E_test', 'RMSE_F_test', 'RMSE_V_test']
    assert columns_check.all()
    assert isinstance(loss.index[0], int)
    assert loss.index[0] == 100
    assert len(loss) == 95


def test_read_loss_single_row(tmpdir):
    """Tries to read a loss.out file that has only a single row"""
    p = tmpdir.join('loss.out')
    p.write('100 2 3 4 5 6 7 8 9 10\n')
    loss = read_loss(f'{p}')
    columns_check = loss.columns == ['total_loss', 'L1', 'L2',
                                     'RMSE_E_train', 'RMSE_F_train', 'RMSE_V_train',
                                     'RMSE_E_test', 'RMSE_F_test', 'RMSE_V_test']
    assert columns_check.all()
    assert isinstance(loss.index[0], int)
    assert loss.index[0] == 100


def test_read_loss_malformed_file(tmpdir):
    """Tries to read a malformed loss.out file"""
    p = tmpdir.join('loss_invalid.out')
    p.write('0 0 0 0 0 0\n')
    with pytest.raises(ValueError):
        read_loss(f'{p}')


# --- read_nep ---
def test_read_nep():
    """Reads a NEP3 nep.txt file"""
    settings, parameters = read_nep('tests/example_files/nep.txt')
    assert settings['zbl'] == (1, 2)
    assert settings['cutoff'] == (8, 4)
    assert settings['n_max'] == (4, 4)
    assert settings['l_max'] == (4, 2, 0)
    assert settings['basis_size'] == (8, 8)
    assert settings['ANN'] == (30, 0)
    assert len(parameters) == 1350


def test_read_nep_malformed_file(tmpdir):
    """Tries to read an invalid nep.in file"""
    p = tmpdir.join('nep.in')
    p.write("""nep123 x x\nnep123 x x\nnep123 x x\nnep123 x x\nnep123 x x\nnep123 x x\n
            nep123 x x\nnep123 x x\n  6.2799901e-02""")
    with pytest.raises(IOError):
        read_nep(f'{p}')


# --- read_kappa ---
def test_read_kappa():
    """Reads a kappa.out file"""
    kappa = read_kappa('tests/example_files/kappa.out')
    columns_check = kappa.columns == ['kx_in', 'kx_out', 'ky_in', 'ky_out',
                                      'kz_tot', 'kx_tot', 'ky_tot']
    assert columns_check.all()
    assert isinstance(kappa.index[0], int)
    assert isinstance(kappa['kx_in'][0], np.float64)
    assert kappa.index[0] == 0


def test_read_kappa_single_row(tmpdir):
    """Reads a single row kappa.out file"""
    p = tmpdir.join('kappa.out')
    p.write('0 1 2 3 4\n')
    kappa = read_kappa(f'{p}')
    columns_check = kappa.columns == ['kx_in', 'kx_out', 'ky_in', 'ky_out',
                                      'kz_tot', 'kx_tot', 'ky_tot']
    assert columns_check.all()
    assert isinstance(kappa.index[0], int)
    assert isinstance(kappa['kx_in'][0], np.float64)
    assert kappa.index[0] == 0


def test_read_kappa_nan():
    """Reads a kappa.out file filed with nans. Should this raise a warning?"""
    kappa = read_kappa('tests/example_files/kappa_nan.out')
    columns_check = kappa.columns == ['kx_in', 'kx_out', 'ky_in', 'ky_out',
                                      'kz_tot', 'kx_tot', 'ky_tot']
    assert columns_check.all()
    assert isinstance(kappa.index[0], int)
    assert isinstance(kappa['kx_in'][0], np.float64)
    assert kappa.index[0] == 0


def test_read_kappa_malformed_file(tmpdir):
    """Tries to read a malformed kappa.out file"""
    p = tmpdir.join('kappa_invalid.out')
    p.write('0 0 0 0 0 0 0 0 0 0 \n')
    with pytest.raises(ValueError):
        read_kappa(f'{p}')


# --- read_hac ---
def test_read_hac():
    """Reads a hac.out file"""
    hac = read_hac('tests/example_files/hac.out')
    columns_check = hac.columns == ['time', 'kx_out', 'ky_out',
                                    'kz_tot', 'kx_tot', 'ky_tot']
    assert columns_check.all()
    assert isinstance(hac.index[0], int)
    assert isinstance(hac['time'][0], np.float64)
    assert hac.index[0] == 0


def test_read_hac_single_row(tmpdir):
    """Reads a single row hac.out file"""
    p = tmpdir.join('hac.out')
    p.write('0 1 2 3 4 5 6 7 8 9 10\n')
    hac = read_hac(f'{p}')
    columns_check = hac.columns == ['time', 'kx_out', 'ky_out',
                                    'kz_tot', 'kx_tot', 'ky_tot']
    assert columns_check.all()
    assert isinstance(hac.index[0], int)
    assert isinstance(hac['time'][0], np.float64)
    assert hac.index[0] == 0


def test_read_hac_nan():
    """Reads a hac.out file filed with nans. Should this raise a warning?"""
    hac = read_hac('tests/example_files/hac_nan.out')
    columns_check = hac.columns == ['time', 'kx_out', 'ky_out',
                                    'kz_tot', 'kx_tot', 'ky_tot']
    assert columns_check.all()
    assert isinstance(hac.index[0], int)
    assert isinstance(hac['time'][0], np.float64)
    assert hac.index[0] == 0


def test_read_hac_malformed_file(tmpdir):
    """Tries to read a malformed hac.out file"""
    p = tmpdir.join('hac_invalid.out')
    p.write('0 0 0 \n')
    with pytest.raises(ValueError):
        read_hac(f'{p}')


# --- read_thermo ---
def test_read_thermo_orthorhombic():
    """Reads a thermo.out file with an orthorhombic structure"""
    thermo = read_thermo('tests/example_files/thermo_ortho_v3.2.out')
    columns_check = thermo.columns == ['temperature', 'kinetic_energy', 'potential_energy',
                                       'stress_xx', 'stress_yy', 'stress_zz',
                                       'cell_xx', 'cell_yy', 'cell_zz']
    assert columns_check.all()


def test_read_thermo_triclinic():
    """Reads a thermo.out file with an triclinic structure"""
    thermo = read_thermo('tests/example_files/thermo_tri_v3.2.out')
    columns_check = thermo.columns == ['temperature', 'kinetic_energy', 'potential_energy',
                                       'stress_xx', 'stress_yy', 'stress_zz',
                                       'cell_xx', 'cell_xy', 'cell_xz',
                                       'cell_yx', 'cell_yy', 'cell_yz',
                                       'cell_zx', 'cell_zy', 'cell_zz'
                                       ]
    assert columns_check.all()


@pytest.mark.parametrize('test_input',
                         [
                             # 9 columns --> orthorhombic cell pre GPUMD v3.3.1
                             (['temperature', 'kinetic_energy', 'potential_energy',
                               'stress_xx', 'stress_yy', 'stress_zz',
                               'cell_xx', 'cell_yy', 'cell_zz']),
                             # 15 columns --> triclinc cell pre GPUMD v3.3.1
                             (['temperature', 'kinetic_energy', 'potential_energy',
                               'stress_xx', 'stress_yy', 'stress_zz',
                               'cell_xx', 'cell_xy', 'cell_xz',
                               'cell_yx', 'cell_yy', 'cell_yz',
                               'cell_zx', 'cell_zy', 'cell_zz']),
                             # 12 columns --> orthorhombic cell GPUMD v3.3.1 forward
                             (['temperature', 'kinetic_energy', 'potential_energy',
                               'stress_xx', 'stress_yy', 'stress_zz',
                               'stress_yz', 'stress_xz', 'stress_xy',
                               'cell_xx', 'cell_yy', 'cell_zz']),
                             # 18 columns --> triclinic cell GPUMD v3.3.1 forward
                             (['temperature', 'kinetic_energy', 'potential_energy',
                               'stress_xx', 'stress_yy', 'stress_zz',
                               'stress_yz', 'stress_xz', 'stress_xy',
                               'cell_xx', 'cell_xy', 'cell_xz',
                               'cell_yx', 'cell_yy', 'cell_yz',
                               'cell_zx', 'cell_zy', 'cell_zz'])
                         ])
def test_read_thermo_pass(test_input):
    """Reads dummy thermo.out files and checks that the correct columns are being returned"""
    s = ' '.join(map(str, range(len(test_input)))) + '\n'
    tmpfile = tempfile.NamedTemporaryFile()
    with open(tmpfile.name, 'w') as f:
        for _ in range(10):
            f.write(s)
    with open(tmpfile.name, 'r') as f:
        thermo = read_thermo(f.name)
    tmpfile.close()
    columns_check = thermo.columns == test_input
    assert columns_check.all()


def test_read_thermo_fail():
    """Checks that ValueError is raised if the number of columns is incorrect"""
    for n in range(1, 20):
        s = ' '.join(map(str, range(n))) + '\n'
        tmpfile = tempfile.NamedTemporaryFile()
        with open(tmpfile.name, 'w') as f:
            for _ in range(10):
                f.write(s)
        with open(tmpfile.name, 'r') as f:
            if n in [9, 15, 12, 18]:
                _ = read_thermo(f.name)
            else:
                with pytest.raises(ValueError):
                    read_thermo(f.name)
        tmpfile.close()


def test_read_thermo_malformed_file(tmpdir):
    """Tries to read a malformed thermo.out file"""
    p = tmpdir.join('thermo_invalid.out')
    p.write('NaN NaN NaN NaN NaN NaN\n')
    with pytest.raises(ValueError):
        read_thermo(f'{p}')


# --- read_xyz and write_xyz ---
def test_write_read_xyz(tmpdir):
    """Writes and reads xyz files"""

    use_triclinic = True

    f = tmpdir.join('atoms.xyz')
    structure_orig = bulk('C').repeat(3)
    write_xyz(f, structure_orig)
    res = read_xyz(f)
    assert isinstance(res, tuple)

    structure_read, settings_read = res
    assert len(structure_orig) == len(structure_read)
    assert np.allclose(structure_orig.cell, structure_read.cell)
    assert np.allclose(structure_orig.positions, structure_read.positions)
    assert isinstance(settings_read, dict)

    cutoff = 9.0
    maximum_neighbors = 99
    write_xyz(f, structure_orig, cutoff=cutoff,
              maximum_neighbors=maximum_neighbors, use_triclinic=use_triclinic)
    res = read_xyz(f)
    structure_read, settings_read = res

    assert np.isclose(settings_read['cutoff'], cutoff)
    assert settings_read['maximum_neighbors'] == maximum_neighbors
    assert settings_read['use_triclinic'] == use_triclinic
    assert not settings_read['has_velocity']

    cutoff = 9.0
    write_xyz(f, structure_orig, cutoff=cutoff)
    res = read_xyz(f)
    structure_read, settings_read = res

    assert np.isclose(settings_read['cutoff'], cutoff)
    assert settings_read['maximum_neighbors'] == 1024
    assert settings_read['use_triclinic']
    assert not settings_read['has_velocity']


def test_write_read_xyz_orthorhombic_has_velocity(tmpdir):
    """Writes and reads an orthorhombic structure with velocity"""
    use_triclinic = False
    f = tmpdir.join('atoms.xyz')
    structure = Atoms('CCC', positions=[(0, 0, 0), (0, 0, 1.1), (0, 0, 2.2)], cell=[1, 2, 3])
    structure.set_velocities([[0, 0, 100], [0, 0, 150], [0, 0, 200]])
    write_xyz(f, structure, use_triclinic=use_triclinic)
    res = read_xyz(f)

    _, settings_read = res
    assert not settings_read['use_triclinic']
    assert settings_read['has_velocity']


def test_write_read_xyz_groupings(tmpdir):
    """Writes and reads a structure with groups"""
    f = tmpdir.join('atoms.xyz')
    structure = Atoms('CCCC',
                      positions=[(0, 0, 0), (0, 0, 1.1), (0, 0, 2.2), (0, 0, 3.3)],
                      cell=[1, 2, 3])
    write_xyz(f, structure, groupings=[[[0, 1], [2, 3]], [[0], [1, 2, 3]]])
    res = read_xyz(f)

    _, settings_read = res
    assert settings_read['number_of_groups'] == 2


def test_write_xyz_invalid_groupings(tmpdir):
    """Tries to write with invalid groups"""
    f = tmpdir.join('atoms.xyz')
    structure = Atoms('CCC', positions=[(0, 0, 0), (0, 0, 1.1), (0, 0, 2.2)], cell=[1, 2, 3])
    with pytest.raises(ValueError):
        # Too many groupings
        write_xyz(f, structure, groupings=[[[0, 1], [2]], [
                  [0, 1], [2]], [[0, 1], [2]], [[0, 1], [2]]])
    with pytest.raises(ValueError):
        # Number of atoms do not add up to the total
        write_xyz(f, structure, groupings=[[[0, 1], [1, 2]]])


# --- Check dump.xyz ---
# Not a unit test, but a good FYI test for our sake
def test_ase_correctly_parses_dump(tmpdir):
    """Check that ASE can correctly read dump files"""
    dump_file = 'tests/example_files/md_no_velocities_or_forces/dump.xyz'
    snapshots = read(dump_file, index=':')
    traj = tmpdir.join('lmao.traj')
    write(f'{traj}', snapshots)  # Make sure that writing works without crashing
    assert all(snapshots[-1][-1].position == [3.10105816, 2.94360168, 2.68268773])


def test_ase_correctly_parses_dump_forces_and_velocities(tmpdir):
    """Check that ASE can correctly read dump files with forces and velocities"""
    dump_file = 'tests/example_files/md_velocities_and_forces/dump.xyz'
    snapshots = read(dump_file, index=':')
    traj = tmpdir.join('lmao.traj')
    write(f'{traj}', snapshots)  # Make sure that writing works without crashing
    assert all(snapshots[-1][-1].position == [3.24197501, 2.84584855, 2.89100032])
    assert all(snapshots[-1].get_array('vel')[-1] == [-0.00033243, 0.00061451, 0.00044743])
    assert all(snapshots[-1].get_array('forces')[-1] == [-0.96089202, -0.01649800, -0.11494368])



import os
import contextlib
from typing import List
import numpy as np
from ase import Atoms
from ase.calculators.calculator import Calculator, all_changes
from tempfile import TemporaryFile
import _nepy


class CPUNEP(Calculator):
    """This class provides an ASE calculator for `nep_cpu`,
    the in-memory CPU implementation of GPUMD.

    Parameters
    ----------
    potential_filename : str
        Path to nep.txt potential
    atoms : Atoms
        Atoms to attach the calculator to
    label : str
        Label for this calculator
    debug : bool, optional
        Flag to toggle debug mode. Prints GPUMD output. Defaults to False.

    Raises
    ------
    FileNotFoundError
        Raises FileNotFoundError if `potential_filename` does not point to a valid file.
    ValueError
        Raises ValueError atoms are not defined when trying to get energies and forces.
    Example
    -------

    >>> calc = CPUNEP('nep.txt')
    >>> atoms.calc = calc
    >>> atoms.get_potential_energy()
    """

    implemented_properties = ['energy', 'energies', 'forces']
    debug = False
    nepy = None

    def __init__(self,
                 potential_filename: str,
                 atoms: Atoms = None,
                 label: str = None,
                 debug: bool = False):

        self.debug = debug
        if not os.path.exists(potential_filename):
            raise FileNotFoundError(f'{potential_filename} does not exist.')
        self.potential_filename = potential_filename

        # Initialize atoms, results and nepy - note that this is also done in Calculator.__init__()
        self.set_atoms(atoms)
        Calculator.__init__(self, label=label, atoms=atoms)
        if atoms is not None:
            self._setup_nepy()

    def _setup_nepy(self):
        """
        Creates an instance of the NEPY class and attaches it to the calculator object.
        The output from `nep.cpp` is only written to STDOUT if debug == True
        """
        if self.atoms is None:
            raise ValueError('Atoms must be defined to get energies and forces.')
        if self.atoms.cell.rank == 0:
            raise ValueError('Atoms must have a defined cell.')

        N_atoms = len(self.atoms)
        self.N_atoms = N_atoms
        c = self.atoms.get_cell(complete=True).flatten()
        cell = [c[0], c[3], c[6], c[1], c[4], c[7], c[2], c[5], c[8]]
        symbols = self.atoms.get_chemical_symbols()
        positions = list(
            self.atoms.get_positions().T.flatten())  # [x1, ..., xN, y1, ... yN,...]

        # Disable output from C++ code by default
        if self.debug:
            self.nepy = _nepy.NEPY(self.potential_filename, self.N_atoms,
                                   cell, symbols, positions)
        else:
            with TemporaryFile('w') as f:
                with contextlib.redirect_stdout(f):
                    self.nepy = _nepy.NEPY(self.potential_filename,
                                           self.N_atoms, cell, symbols, positions)

    def set_atoms(self, atoms: Atoms):
        """Updates the Atoms object.

        Parameters
        ----------
        atoms : Atoms
            Atoms to attach the calculator to
        """
        self.atoms = atoms
        self.results = {}
        self.nepy = None

    def _update_symbols(self):
        """Update atom symbols in NEPY.
        """
        symbols = self.atoms.get_chemical_symbols()
        self.nepy.set_symbols(symbols)

    def _update_cell(self):
        """Update cell parameters in NEPY.
        """
        c = self.atoms.get_cell(complete=True).flatten()
        cell = [c[0], c[3], c[6], c[1], c[4], c[7], c[2], c[5], c[8]]
        self.nepy.set_cell(cell)

    def _update_positions(self):
        """Update atom positions in NEPY.
        """
        positions = list(
            self.atoms.get_positions().T.flatten())  # [x1, ..., xN, y1, ... yN,...]
        self.nepy.set_positions(positions)

    def calculate(self,
                  atoms: Atoms = None,
                  properties: List[str] = None,
                  system_changes: List[str] = all_changes):
        """Calculate energy, per atom energies and forces.

        Parameters
        ----------
        atoms : Atoms, optional
            System for which to calculate properties, by default None
        properties : List[str], optional
            Properties to calculate, by default None
        system_changes : List[str], optional
            Changes to the system since last call, by default all_changes
        """
        if properties is None:
            properties = self.implemented_properties

        Calculator.calculate(self, atoms, properties, system_changes)

        if self.nepy is None:
            # Create new NEPY interface
            self._setup_nepy()
        # Update existing NEPY interface
        for change in system_changes:
            if change == 'positions':
                self._update_positions()
            elif change == 'numbers':
                self._update_symbols()
            elif change == 'cell':
                self._update_cell()

        energies, forces, _ = self.nepy.get_potential_forces_and_virials()
        energies_per_atom = np.array(energies)
        energy = energies_per_atom.sum()
        forces_per_atom = np.array(forces).reshape(-1, self.N_atoms).T
        # virials_per_atom = np.array(virials).reshape(-1, self.N_atoms).T

        self.results['energy'] = energy
        self.results['energies'] = energies_per_atom
        self.results['forces'] = forces_per_atom



import pytest

from calorine.data_analysis import (analyze_data,
                                    get_autocorrelation_function,
                                    get_correlation_length,
                                    get_error_estimate,
                                    _estimate_correlation_length_from_acf,
                                    _estimate_error)
import numpy as np


@pytest.fixture
def data():
    """Simplified but representative data series."""
    return np.arange(1000) % 20


@pytest.fixture
def data_flat():
    """Data series without variation."""
    return np.zeros(20)


@pytest.fixture
def target_values():
    """Target values for reasonable data series."""
    return {'correlation_length': 4,
            'error_estimate': 0.7156495464548879,
            'mean': 9.5,
            'std': 5.766281297335398}


@pytest.fixture
def target_values_flat():
    """Target values for data series without any variation."""
    return {'correlation_length': np.nan,
            'error_estimate': np.nan,
            'mean': 0,
            'std': 0}


def test_analyze_data(data, target_values):
    """Tests the analyze_data function."""
    res = analyze_data(data)
    assert isinstance(res, dict)
    assert len(res) == len(target_values)
    for key, val in target_values.items():
        assert key in res
        assert abs(res[key] - val) < 1e-6


def test_analyze_data_flat(data_flat, target_values_flat):
    """Tests the analyze_data function."""
    res = analyze_data(data_flat)
    assert isinstance(res, dict)
    assert len(res) == len(target_values_flat)
    for key, val in target_values_flat.items():
        assert key in res
        if np.isnan(val):
            assert np.isnan(res[key])
        else:
            assert abs(res[key] - val) < 1e-6


def test_get_autocorrelation_function(data):
    """Tests the calculation of autocorrelation function."""
    autocorr = get_autocorrelation_function(data)
    assert isinstance(autocorr, np.ndarray)
    assert len(autocorr) == len(data) - 1
    assert abs(max(autocorr) - 1) < 1e-6
    assert abs(min(autocorr) - -0.5037974683544306) < 1e-6
    for i in range(len(autocorr) // 20):
        assert abs(autocorr[20 * i] - 1) < 1e-6


def test_get_autocorrelation_function_flat(data_flat):
    """Tests the calculation of autocorrelation function with a flat dataset."""
    autocorr = get_autocorrelation_function(data_flat)
    assert isinstance(autocorr, np.ndarray)
    assert len(autocorr) == len(data_flat) - 1
    for i in autocorr:
        assert np.isnan(i)


def test_get_autocorrelation_function_max_lag(data):
    """Tests the calculation of autocorrelation function with max_lag specified."""
    max_lag = 40
    autocorr = get_autocorrelation_function(data, max_lag=max_lag)
    assert isinstance(autocorr, np.ndarray)
    assert len(autocorr) == max_lag
    for i in range(len(autocorr) // 20):
        assert abs(autocorr[20 * i] - 1) < 1e-6


def test_get_autocorrelation_function_bad_max_lag(data):
    """Tests the calculation of autocorrelation function with too long max_lag."""
    max_lag = len(data) + 1
    with pytest.raises(ValueError) as e:
        get_autocorrelation_function(data, max_lag=max_lag)
    assert 'max_lag should be between 1 and len(data)-1' in str(e)


def test_get_correlation_length(data, target_values):
    """Tests calculation of correlation length."""
    corr_length = get_correlation_length(data)
    assert abs(corr_length - target_values['correlation_length']) < 1e-6


def test_get_correlation_length_flat(data_flat):
    """Tests calculation of correlation length with flat data series."""
    corr_length = get_correlation_length(data_flat)
    assert corr_length is None


def test_get_error_estimate(data, target_values):
    """Tests get_error_estimate function."""
    error = get_error_estimate(data)
    assert abs(error - target_values['error_estimate']) < 1e-6


def test_get_error_estimate_flat(data_flat):
    """Tests get_error_estimate function with flat data series."""
    error = get_error_estimate(data_flat)
    assert error is None


def test_estimate_correlation_length_from_acf():
    """Tests calculation of correlation length estimate from autocorrelation function."""
    acf = [1, 0.5, 0.01, -0.01, 1]
    corr_length = _estimate_correlation_length_from_acf(acf)
    assert corr_length == 2


def test_estimate_correlation_length_from_acf_flat_data():
    """
    Tests calculation of correlation length estimate from autocorrelation function
    with a flat data series.
    """
    acf = [1, 1, 1, 1, 1]
    corr_length = _estimate_correlation_length_from_acf(acf)
    assert corr_length is None


def test_estimate_error(data, target_values):
    """Tests error estimate function"""
    error = _estimate_error(data, correlation_length=4, confidence=0.95)
    assert np.abs(error - target_values['error_estimate']) < 1e-6



import os
import contextlib
import shutil
import tempfile
from typing import List, Tuple
import warnings
import numpy as np
import _nepy
from ase import Atoms


def _create_dummy_nep2(potential_filename: str, symbols: List[str]) -> None:
    """Create a dummy NEP2 model, i.e., a model for which there are no descriptor parameters.
    This is to be used when one wants descriptors not pertaining to a specific NEP3 model.

    Parameters
    ----------
    potential_filename
        Path to which the NEP2 model will be saved.
    symbols
        Atomic elements in the configuration for which to compute descriptors.
    """
    unique_symbols = []
    for sym in symbols:
        if sym not in unique_symbols:
            unique_symbols.append(sym)
    with open(potential_filename, 'w') as f:
        f.write(f'nep {len(unique_symbols)} ')
        for symbol in unique_symbols:
            f.write(f'{symbol} ')
        f.write('\n')
        # Write rest of header
        f.write('cutoff 6 4\nn_max 15 8\nl_max 4\nANN 30 0\n')
        # Write dummy parameters
        # The number of parameters in the network is:
        # N_par = (N_des+2)N_neu + 1 + N_typ**2 (n_max^R + n_max^A + 2)
        # Default settings: 1621 + N_typ**2 * 25
        # The last 52 1:s are the normalization parameters for the descriptors.
        for _ in range(1621 + len(unique_symbols)**2 * 25 + 52):
            f.write(f'  {1e0:e}\n')


def _create_tmp_dir(debug: bool) -> str:
    """Create temporary directory.

    Parameters
    ----------
    debug
        Flag that indicates if debugging. Use a hardcoded path when debugging.

    Returns
    -------
    str
         Path to temporary directory
    """
    if debug:
        tmp_dir = './tmp_nepy/'
        # os.mkdir(tmp_dir)
        try:
            os.mkdir(tmp_dir)
            return tmp_dir
        except FileExistsError:
            raise FileExistsError('Please delete or move the conflicting directory.')
    return tempfile.mkdtemp()


def _clean_tmp_dir(directory: str) -> None:
    """Remove temporary directory.

    Parameters
    ----------
    directory
        Path to temporary directory
    """
    shutil.rmtree(directory)


def set_default_cell(structure: Atoms, box_length: float = 100):
    """Adds a cubic box to an Atoms object. Atoms object is edited in-place.

    Parameters
    ----------
    structure
        Structure to add box to
    box_length
        Cubic box side length in Å, by default 100
    """
    structure.set_cell([[box_length, 0, 0], [0, box_length, 0], [0, 0, box_length]])
    structure.center()


def get_descriptors(
        structure: Atoms,
        potential_filename: str = None,
        debug: bool = False) -> np.ndarray:
    """Calculates the NEP descriptors for a given structure. A NEP model defined by a nep.txt
    can additionally be provided to get the NEP3 model specific descriptors. If no model is
    provided, a dummy NEP2 model suitable for the provided structure will be created and used.

    Parameters
    ----------
    structure
        Input structure
    potential_filename
        Path to NEP potential. Defaults to None.
    debug
        Flag to toggle debug mode. Makes the generated dummy NEP2 model available
        in a local tmp directory, as well as prints GPUMD output. Defaults to False.

    Returns
    -------
        Descriptors for the supplied structure, with shape (N_atoms, descriptor components)
    """
    local_structure = structure.copy()
    if local_structure.cell.rank == 0:
        warnings.warn('Using default unit cell (cubic with side 100 Å).')
        set_default_cell(local_structure)

    N_atoms = len(local_structure)
    c = local_structure.get_cell(complete=True).flatten()
    cell = [c[0], c[3], c[6], c[1], c[4], c[7], c[2], c[5], c[8]]

    symbols = local_structure.get_chemical_symbols()
    positions = list(local_structure.get_positions().T.flatten())  # [x1, ..., xN, y1, ... yN,...]

    use_dummy_nep2_potential = potential_filename is None
    if use_dummy_nep2_potential:
        tmp_dir = _create_tmp_dir(debug)
        potential_filename = f'{tmp_dir}/nep.txt'
        _create_dummy_nep2(potential_filename, symbols)

    # Disable output from C++ code by default
    if debug:
        nepy = _nepy.NEPY(potential_filename, N_atoms, cell, symbols, positions)
    else:
        with open(os.devnull, 'w') as f:
            with contextlib.redirect_stdout(f):
                with contextlib.redirect_stderr(f):
                    nepy = _nepy.NEPY(potential_filename, N_atoms, cell, symbols, positions)

    all_descriptors = nepy.get_descriptors()
    descriptors_per_atom = np.array(all_descriptors).reshape(-1, N_atoms).T

    if use_dummy_nep2_potential and not debug:
        _clean_tmp_dir(tmp_dir)
    if use_dummy_nep2_potential and debug:
        print(f'DEBUG - Directory containing dummy nep.in: {tmp_dir}')
    return descriptors_per_atom


def get_potential_forces_and_virials(structure: Atoms,
                                     potential_filename: str = None,
                                     debug: bool = False) -> Tuple[np.ndarray]:
    """Calculates the per-atom potential, forces and virials for a given structure.
    A NEP model defined by a nep.txt-file needs to be provided.

    Parameters
    ----------
    structure
        Input structure
    potential_filename
        Path to NEP potential. Defaults to None.
    debug
        Flag to toggle debug mode. Prints GPUMD output. Defaults to False.

    Returns
    -------
        (potential, forces, virials) for the supplied structure,
        with shapes (N_atoms,), (N_atoms,3), (N_atoms,9)
    """
    if potential_filename is None:
        raise ValueError('Potential must be defined!')
    local_structure = structure.copy()
    if local_structure.cell.rank == 0:
        warnings.warn('Using default unit cell (cubic with side 100 Å).')
        set_default_cell(local_structure)

    N_atoms = len(local_structure)
    c = local_structure.get_cell(complete=True).flatten()
    cell = [c[0], c[3], c[6], c[1], c[4], c[7], c[2], c[5], c[8]]

    symbols = local_structure.get_chemical_symbols()
    positions = list(local_structure.get_positions().T.flatten())  # [x1, ..., xN, y1, ... yN,...]

    # Disable output from C++ code by default
    if debug:
        nepy = _nepy.NEPY(potential_filename, N_atoms, cell, symbols, positions)
    else:
        with open(os.devnull, 'w') as f:
            with contextlib.redirect_stdout(f):
                with contextlib.redirect_stderr(f):
                    nepy = _nepy.NEPY(potential_filename, N_atoms, cell, symbols, positions)

    energies, forces, virials = nepy.get_potential_forces_and_virials()
    forces_per_atom = np.array(forces).reshape(-1, N_atoms).T
    virials_per_atom = np.array(virials).reshape(-1, N_atoms).T
    return np.array(energies), forces_per_atom, virials_per_atom



# -*- coding: utf-8 -*-
"""
Submodule of calorine containing `nep_cpu` interface.
"""

from .nep import get_descriptors, get_potential_forces_and_virials
from .cpunep_calculator import CPUNEP

__all__ = ['CPUNEP',
           'get_descriptors',
           'get_potential_forces_and_virials']



import pytest
import contextlib
import numpy as np
from ase import Atoms
from calorine.nepy import CPUNEP
from calorine import GPUNEP

vacuum_cell = ([100, 0, 0], [0, 100, 0], [0, 0, 100])
PbTe = Atoms('TePb', positions=[(0, 0, 0), (0, 0.0, 1.1)],
             cell=vacuum_cell)
C = Atoms('C', positions=[(0, 0, 0)],
          cell=vacuum_cell)
CO = Atoms('CO', positions=[(0, 0, 0), (0, 0.0, 1.1)],
           cell=vacuum_cell)
CC = Atoms('CC', positions=[(0, 0, 0), (0, 0.0, 1.1)],
           cell=vacuum_cell)
CON = Atoms('CON', positions=[(0, 0, 0), (0, 0.0, 1.1), (0, 0.0, 2.2)],
            cell=vacuum_cell)


# --- get_potential_forces_and_virials ---
def test_get_potential_and_forces_NEP3():
    """NEP3 model supplied. Compares results to output from `nep_cpu`"""
    nep3 = 'tests/nep_models/PbTe_NEP3.txt'
    calc = CPUNEP(nep3)
    PbTe.calc = calc
    energy = PbTe.get_potential_energy()
    forces = PbTe.get_forces()

    assert energy.shape == ()
    assert forces.shape == (2, 3)
    assert np.allclose(forces[0, :], -forces[1, :], atol=1e-12, rtol=0)  # Newton III
    expected_forces = np.loadtxt(
        'tests/expected_descriptors/PbTe_2atom_NEP3_force.out')
    assert np.allclose(forces, expected_forces, atol=1e-12, rtol=0)


def test_get_potential_and_forces_set_atoms_constructor():
    """Set atoms directly when creating the calculator"""
    nep3 = 'tests/nep_models/PbTe_NEP3.txt'
    CPUNEP(nep3, atoms=PbTe)
    energy = PbTe.get_potential_energy()
    forces = PbTe.get_forces()

    assert energy.shape == ()
    assert forces.shape == (2, 3)
    assert np.allclose(forces[0, :], -forces[1, :], atol=1e-12, rtol=0)  # Newton III
    expected_forces = np.loadtxt(
        'tests/expected_descriptors/PbTe_2atom_NEP3_force.out')
    assert np.allclose(forces, expected_forces, atol=1e-12, rtol=0)


def test_get_potential_and_forces_set_atoms_calculate():
    """Set atoms directly when calling calculate on the calculator"""
    nep3 = 'tests/nep_models/PbTe_NEP3.txt'
    calc = CPUNEP(nep3)
    calc.calculate(atoms=PbTe)
    results = calc.results
    energy, energies, forces = results['energy'], results['energies'], results['forces']

    assert energy.shape == ()
    assert energies.shape == (2,)  # Per atom (makes no physical sense)
    assert forces.shape == (2, 3)
    assert np.allclose(forces[0, :], -forces[1, :], atol=1e-12, rtol=0)  # Newton III
    expected_forces = np.loadtxt(
        'tests/expected_descriptors/PbTe_2atom_NEP3_force.out')
    assert np.allclose(forces, expected_forces, atol=1e-12, rtol=0)


def test_get_potential_forces_and_virials_NEP3_debug(tmpdir):
    """Compares result with debug flag enabled."""
    nep3 = 'tests/nep_models/PbTe_NEP3.txt'
    calc = CPUNEP(nep3, debug=True)
    p = tmpdir.join('nep_out.tmp')
    with open(p, 'w') as f:
        with contextlib.redirect_stdout(f):
            with contextlib.redirect_stderr(f):
                PbTe.calc = calc
                PbTe.get_potential_energy()
                PbTe.get_forces()
    with open(p, 'r') as f:
        lines = p.readlines()
        print(lines)
        assert lines[0] == 'Use the NEP3 potential with 2 atom type(s).\n'
        assert len(lines) == 16


def test_get_potential_and_forces_dummy_NEP2():
    """Dummy NEP2 model supplied. Compares results to output from `nep_cpu` for another system"""
    nep2 = 'tests/nep_models/CO_NEP2_dummy.txt'
    calc = CPUNEP(nep2)
    CO.calc = calc
    energy = CO.get_potential_energy()
    forces = CO.get_forces()

    assert energy.shape == ()
    assert forces.shape == (2, 3)
    assert np.allclose(forces[0, :], -forces[1, :], atol=1e-12, rtol=0)  # Newton III
    expected_forces = np.loadtxt(
        'tests/expected_descriptors/CO_2atom_NEP2_dummy_force.out')
    assert np.allclose(forces, expected_forces, atol=1e-12, rtol=0)


def test_get_potential_and_forces_several_different_species():
    """Check that forces are correct for a CON system.
    Note that these forces should be exactly zero for this system
    since the NEP2 dummy potential treats all atom species as identical atm."""
    nep2 = 'tests/nep_models/CON_NEP2_dummy.txt'
    calc = CPUNEP(nep2)
    CON.calc = calc
    energy = CON.get_potential_energy()
    forces = CON.get_forces()

    assert energy.shape == ()
    assert forces.shape == (3, 3)
    assert np.allclose(forces[0, :], -forces[1, :], atol=1e-12, rtol=0)  # Newton III
    expected_forces = np.loadtxt(
        'tests/expected_descriptors/CON_3atom_NEP2_dummy_force.out')
    assert np.allclose(forces, expected_forces, atol=1e-12, rtol=0)
    assert np.allclose(forces, np.zeros((3, 3)), atol=1e-12, rtol=0)


def test_get_potential_and_forces_dummy_NEP2_independent_of_species():
    """Dummy NEP2 energies and forces should be independent of atom species"""
    nep2 = 'tests/nep_models/PbTe_NEP2_dummy.txt'
    calc = CPUNEP(nep2)
    PbTe.calc = calc
    PbTe_energy = PbTe.get_potential_energy()
    PbTe_forces = PbTe.get_forces()

    nep2 = 'tests/nep_models/CO_NEP2_dummy.txt'
    calc = CPUNEP(nep2)
    CO.calc = calc
    CO_energy = CO.get_potential_energy()
    CO_forces = CO.get_forces()

    expected_forces_CO = np.loadtxt(
        'tests/expected_descriptors/CO_2atom_NEP2_dummy_force.out')

    assert np.allclose(PbTe_energy, CO_energy, atol=1e-12, rtol=0)
    assert np.allclose(PbTe_forces, CO_forces, atol=1e-12, rtol=0)
    assert np.allclose(CO_forces, expected_forces_CO, atol=1e-12, rtol=0)


def test_get_potential_and_forces_update_positions():
    """Update the positions and make sure that the energies and forces are also updated"""
    nep3 = 'tests/nep_models/PbTe_NEP3.txt'
    calc = CPUNEP(nep3)

    copy = PbTe.copy()
    copy.calc = calc
    energy_initial = copy.get_potential_energy()
    forces_initial = copy.get_forces()

    # Move atoms slightly
    copy.set_positions([[0, 0, 0], [0, 0, 2.2]])
    energy_after = copy.get_potential_energy()
    forces_after = copy.get_forces()

    diff_energy = np.abs(energy_after - energy_initial)
    diff_force = forces_initial - forces_after
    assert np.isclose(diff_energy, 1.80751674, atol=1e-8, rtol=0)
    assert np.allclose(diff_force, [[0, 0, 4.65672972], [0, 0, -4.65672972]], atol=1e-8, rtol=0)


def test_get_potential_and_forces_update_cell():
    """Update the cell and make sure that the energies and forces are still the same"""
    nep3 = 'tests/nep_models/PbTe_NEP3.txt'
    calc = CPUNEP(nep3)

    copy = PbTe.copy()
    copy.calc = calc
    energy_initial = copy.get_potential_energy()
    forces_initial = copy.get_forces()

    # Change box
    copy.set_cell([[20, 0, 0], [0, 20, 0], [0, 0, 20]], scale_atoms=True)

    energy_after = copy.get_potential_energy()
    forces_after = copy.get_forces()

    diff_energy = np.abs(energy_after - energy_initial)
    diff_force = forces_initial - forces_after
    assert np.isclose(diff_energy, 0.72005750, atol=1e-8, rtol=0)
    assert np.allclose(diff_force, [[0, 0, 6.58735098], [0, 0, -6.58735098]], atol=1e-8, rtol=0)


def test_get_potential_and_forces_update_numbers():
    """Update the atom numbers (species) and make sure that the
    energies and forces are also updated
    """
    nep3 = 'tests/nep_models/PbTe_NEP3.txt'
    calc = CPUNEP(nep3)

    copy = PbTe.copy()
    copy.calc = calc
    energy_initial = copy.get_potential_energy()
    forces_initial = copy.get_forces()

    # Change atomic numbers
    copy.set_atomic_numbers([82, 82])  # Pb_2

    energy_after = copy.get_potential_energy()
    forces_after = copy.get_forces()

    diff_energy = np.abs(energy_after - energy_initial)
    diff_force = forces_initial - forces_after
    assert np.isclose(diff_energy, 1.86577361, atol=1e-8, rtol=0)
    assert np.allclose(diff_force, [[0, 0, 1.07038059], [0, 0, -1.07038059]], atol=1e-8, rtol=0)


def test_reset_calculator_on_atoms_change():
    """Reset the calculator when changing the system.
    """
    nep3 = 'tests/nep_models/PbTe_NEP3.txt'
    calc = CPUNEP(nep3)

    copy = PbTe.copy()
    copy.calc = calc
    energy_initial = copy.get_potential_energy()
    forces_initial = copy.get_forces()

    # Copy system
    atoms_copy = copy.copy()
    original_cell = copy.cell.copy()
    atoms_copy.calc = calc

    assert calc.results == {}
    assert calc.nepy is None

    # Scale cell
    atoms_copy.set_cell(1.1*original_cell, scale_atoms=True)
    energy_after = atoms_copy.get_potential_energy()
    forces_after = atoms_copy.get_forces()

    diff_energy = np.abs(energy_after - energy_initial)
    diff_force = forces_initial - forces_after
    assert np.isclose(diff_energy, 0.28572432, atol=1e-8, rtol=0)
    assert np.allclose(diff_force, [[0, 0, 0.27278336], [0, 0, -0.27278336]], atol=1e-8, rtol=0)


def test_get_potential_and_forces_no_cell():
    """Should raise error if no cell is supplied"""
    nep3 = 'tests/nep_models/PbTe_NEP3.txt'
    calc = CPUNEP(nep3)
    atoms = Atoms('C', positions=[(0, 0, 0)])
    atoms.calc = calc
    with pytest.raises(ValueError)as e:
        atoms.get_potential_energy()
    assert 'Atoms must have a defined cell.' in str(e)


def test_get_potential_and_forces_no_potential():
    """Tries to get potentials and forces without specifying potential"""
    with pytest.raises(FileNotFoundError)as e:
        CPUNEP('nep.txt')
    assert 'nep.txt does not exist.' in str(e)


def test_get_potential_and_forces_no_atoms():
    """Tries to get potential and forces without specifying atoms"""
    nep3 = 'tests/nep_models/PbTe_NEP3.txt'
    calc = CPUNEP(nep3)
    with pytest.raises(ValueError)as e:
        calc.calculate()
    assert 'Atoms must be defined to get energies and forces.' in str(e)


def test_CPU_GPU_equivalent():
    """Assert that the CPU and GPU implementation are equivalent
    """
    nep3 = 'tests/nep_models/PbTe_NEP3.txt'
    # CPU
    PbTe.calc = CPUNEP(nep3)
    cpu_energy = PbTe.get_potential_energy()
    cpu_forces = PbTe.get_forces()
    # GPU
    PbTe.calc = GPUNEP(nep3)
    gpu_energy = PbTe.get_potential_energy()
    gpu_forces = PbTe.get_forces()
    assert np.isclose(cpu_energy, gpu_energy, atol=1e-6)
    assert np.allclose(cpu_forces, gpu_forces, atol=1e-6)
